/*
  Copyright (C) 2004 Openfuel (Pty) Ltd

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  Author     : Marc Welz

  Description: This program bridges ethernet frames from a connexant 
               USB device to a linux tap device
 
  Credits    : Uses code fragments from the connexant accessrunner project
               GPL (C) 2002, 2003 Josep Comas

  Problems   : io to the device via USB seems quite flaky, it was written
               without formal documentation
               - endpoint 0x01/81 drops/duplicates packets 
               - endpoint 0x02/82 seems to append occasional garbage bytes 
*/

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <termios.h>
#include <unistd.h>
#include <signal.h>
#include <syslog.h>

#include <usb.h>
#include <pthread.h>

#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <linux/if.h>
#include <linux/if_tun.h>

#define NAME       16
#define TIMEOUT  2000
#define BUFFER   2048
#define RESEND      2

#define TAPUSB_ETHEP     0x02
#define TAPUSB_CTLEP     0x01
#define TAPUSB_VENDOR  0x0572
#define TAPUSB_PRODUCT 0xcb06
#define TAPUSB_CONFIG       1
#define TAPUSB_INTERFACE    0
#define TAPUSB_BUFFER      64

#define TAPUSB_MAXSEC       8
#define TAPUSB_ARBSECM    500

struct usb_dev_handle *usbah;
int usbeth;
int usbctl;
int usbbuf;

/* WARNING: should be guarded by locks */
volatile int verbose = 1; 
volatile int tapfd = (-1);

/* accounting detail */
volatile int subwe = 0;
volatile int subwc = 0;
volatile int subre = 0;
volatile int subrc = 0;

void handle(int number)
{
  if(tapfd >= 0){
    close(tapfd);
    tapfd = (-1);
  }
}

void printbuffer(unsigned char *buffer, int len, int width)
{
  int q, i, j;

  q = (width / 4) - 3;

  for (i = 0; i < len; i += q) {
    for (j = i; j < len && j < i + q; j++) {
      fprintf(stderr, "%02x ", buffer[j]);
    }
    while (j < i + q) {
      j++;
      fprintf(stderr, "   ");
    }
    fprintf(stderr, "  ");
    for (j = i; j < len && j < i + q; j++) {
      fprintf(stderr, "%c", isprint(buffer[j]) ? buffer[j] : '.');
    }
    fprintf(stderr, "\n");
  }
}

int check(unsigned char *buffer, int len)
{
  if(len <= 14){
    if(verbose){
      fprintf(stderr, "check: short packet: %dbytes\n", len);
    }
    return -1;
  }

  if(buffer[12] != 0x08){
    if(verbose){
      fprintf(stderr, "check: bad ethernet type\n");
    }
    return -1;
  }
  
  switch(buffer[13]){
    case 0x06 :
      return 0;
    case 0x00 :
      if((buffer[14] & 0xf0) != 0x40){
        if(verbose){
          fprintf(stderr, "check: not an ipv4 packet\n");
        }
        return -1;
      }
      return 0;
    default : 
      if(verbose){
        fprintf(stderr, "check: unknown ethernet type\n");
      }
      return -1;
  }

  return -1;
}

void online()
{
  printf("status:  print status information\n");
  printf("help:    this help\n");
  printf("exit:    terminate\n");
  printf("verbose: increase verbosity\n");
  printf("quiet:   reduce verbosity\n");
}

void prompt()
{
  printf("%d> ", verbose);
  fflush(stdout);
}

struct usb_dev_handle *sub_open(int vendor, int product, int config, int inter)
{
  int devc, busc;
  struct usb_bus *bus;
  struct usb_device *dev;
  struct usb_dev_handle *handle;
  int error;

  usb_init();

  busc = usb_find_busses();
  if(busc < 0){
    fprintf(stderr, "unable to find busses\n");
    return NULL;
  }

  devc = usb_find_devices();
  if(devc < 0){
    fprintf(stderr, "unable to find devices\n");
    return NULL;
  }

  if(verbose){
    fprintf(stderr, "found %d devices on %d bus%s\n", devc, busc, (busc == 1) ? "" : "ses");
  }

  bus = usb_get_busses();
  if(bus == NULL){
    return NULL;
  }

  for(dev = bus->devices; dev;){
    if(dev->descriptor.idVendor == vendor && dev->descriptor.idProduct == product){
      if(verbose){
        fprintf(stderr, "type=0x%02x\n", dev->config->bDescriptorType);
        fprintf(stderr, "interfaces=0x%02x\n", dev->config->bNumInterfaces);
        fprintf(stderr, "configuration=0x%02x/0x%02x\n", dev->config->iConfiguration, dev->config->bConfigurationValue);
        fprintf(stderr, "power=0x%x\n", dev->config->MaxPower);
      }
      
      handle = usb_open(dev);
      if(handle){
#if 0
        error = usb_set_configuration(handle, config);
        if(error >= 0){
#endif
          error = usb_claim_interface(handle, inter);
          if(error >= 0){
            return handle;
          }
          if(verbose){
            fprintf(stderr, "unable to claim interface: %s\n", strerror(errno));
          }
#if 0
        } 
#endif
        if(verbose){
          fprintf(stderr, "unable to set configuration %d for 0x%x:0x%x: %s\n", config, vendor, product, usb_strerror());
          fprintf(stderr, "resetting device\n");
        }
        /* WARNING: not sure if a reset is advisable */
        usb_reset(handle);
        usb_close(handle);
      } 
      if(verbose){
        fprintf(stderr, "unable to open a matching device\n");
      }
    }

    if(dev->next){
      dev = dev->next;
    } else {
      bus = bus->next;
      if(bus){
        dev = bus->devices;
      } else {
        return NULL;
      }
    }
  }

  return NULL;
}

int sub_close(struct usb_dev_handle *handle, int inter)
{
  usb_release_interface(handle, inter);
  usb_close(handle);

  return 0;
}

int tap_open(char *name)
{
#if 0
  return open("/dev/tap0", O_RDWR);
#else
  int fd;
  struct ifreq ifr;

  fd = open("/dev/net/tun", O_RDWR);
  if(fd < 0){
    return -1;
  }

  memset(&ifr, 0, sizeof(ifr));

  ifr.ifr_flags = IFF_TAP | IFF_NO_PI;
  /* ifr.ifr_flags = IFF_TAP; */

  if(name){
    strncpy(ifr.ifr_name, name, IFNAMSIZ);
    if(verbose){
      fprintf(stderr, "tap: interface %s\n", name);
    }
  }

  if (ioctl(fd, TUNSETIFF, (void *) &ifr) < 0) {
    return -1;
  }

  return fd;
#endif
}

int tap_close(int fd)
{
  return close(fd);
}

int console(void)
{
  unsigned char buffer[BUFFER];
  int rr, tty, run, sr;
  fd_set fsr;
  struct timeval tv;

  tty = isatty(STDIN_FILENO);
  if(tty){
    printf("press ? for help\n");
  } 

  for(run = 1; run > 0;){
    if(tty){
      prompt();
    }

    do{
      FD_ZERO(&fsr);
      FD_SET(STDIN_FILENO, &fsr);

      tv.tv_sec = TAPUSB_MAXSEC;
      tv.tv_usec = 0;

      sr = select(STDIN_FILENO + 1, &fsr, NULL, NULL, &tv);
      if(sr > 0){
        rr = read(STDIN_FILENO, buffer, BUFFER);
      } else {
        rr = 0;
      }
      if(tapfd < 0){
        sr = (-1);
        run = 0;
        rr = 0;
      }
    } while(sr == 0);

    if(rr > 0){
      switch(buffer[0]){
        case '\n' :
        case '\r' :
          break;
        case 's' :
          printf("verbosity=%d, tapfd=%d (%s)\n", verbose, tapfd, (tapfd < 0) ? "broken" : "ok");
          printf("usb data error rates: read=%d/%d, write=%d/%d\n", subre, subrc, subwe, subwc);
          break;
        case 'h' :
        case '?' :
          online();
          fflush(stdout);
          break;
        case 'e' :
        case 'x' :
          run = 0;
          break;
        case 'q' :
          if(verbose > 0){
            verbose--;
          }
          break;
        case 'v' :
          verbose++;
          break;
        default :
          fprintf(stderr, "bad command: try help\n");
          break;
      }
    } else {
      if(rr < 0){
        if(verbose){
          fprintf(stderr, "console: read failed: %s\n", strerror(errno));
        }
        run--;
      } else {
        if(verbose){
          fprintf(stderr, "console: eof\n");
        }
        run = 0;
      }
    }
  }

  close(tapfd);
  tapfd = (-1);

  if(tty){
    printf("\n");
  }

  return 0;
}

void *tap2sub(void *data)
{
  unsigned char buffer[BUFFER];
  int sr, rr, wr, hw, sw, run, full;
  fd_set fsr;
  struct timeval tv;

  for(run = 1; run > 0;){

    tv.tv_sec = TAPUSB_MAXSEC;
    tv.tv_usec = 0;

    if(tapfd < 0){
      run = 0;
      sr = 0;
    } else {
      FD_ZERO(&fsr);
      FD_SET(tapfd, &fsr);
      sr = select(tapfd + 1, &fsr, NULL, NULL, &tv);
    }
    if(sr > 0){
      rr = read(tapfd, buffer + 2, BUFFER - 4);
      if(rr > 0){
        subwc++;
        if(verbose > 1){
          fprintf(stderr, "---> from tap\n");
          printbuffer(buffer + 2, rr, 70);
        }
        buffer[0] = 0xff & rr;
        buffer[1] = ((rr >> 8) & 0xff);
        rr += 2;
#if 0
        if((rr % usbbuf) == 0){
          buffer[rr++] = 0;
          buffer[rr++] = 0;
        }
#endif
        hw = 0;
        for(full = 1; full;){
          sw = (rr - hw);
          if(sw >= usbbuf){
            sw = usbbuf;
          } else {
            full = 0;
          }
          if(verbose > 2){
            fprintf(stderr, "---> to usb\n");
            printbuffer(buffer + hw, sw, 70);
          }
          wr = usb_bulk_write(usbah, usbeth, buffer + hw, sw, TAPUSB_ARBSECM);
          if(wr < sw){
            if(verbose){
              fprintf(stderr, "usb: write failed: %s\n", (wr < 0) ? usb_strerror() : "partial data");
            }
            full = 0;
            subwe++;
          } else {
            hw += sw;
          }
        }
      } else {
        if(verbose){
          fprintf(stderr, "tap: read failed: %s\n", rr ? strerror(errno) : "eof");
        }
        run = 0;
      }
    }
  }

  if(tapfd >= 0){
    close(tapfd);
    tapfd = (-1);
  }

  return NULL;
}

void *sub2tap(void *data)
{
  unsigned char buffer[BUFFER];
  int rr, hr, wr, sw, run;

  hr = 0;

  for(run = 1; run > 0;){
    if(tapfd >= 0){
      if((hr + usbbuf) >= BUFFER){
        hr = 0;
      }
      rr = usb_bulk_read(usbah, usbeth | 0x80, buffer + hr, usbbuf, TAPUSB_MAXSEC * 1000);
    } else {
      if(verbose){
        fprintf(stderr, "usb: saw eof\n");
      }
      rr = (-1);
      run = 0;
    }
    /* WARNING: occasionally we receive garbage. Unclear if this has
     * to do with an alignment issue (append a byte if size odd ?)
     */
    if(rr > 0){
      if(verbose > 1){
        fprintf(stderr, "---> from usb: %d+%d\n", hr, rr);
        printbuffer(buffer + hr, rr, 70);
      }
      hr += rr;
      if((hr < 15) || (buffer[14] != 0x08)){
        if(verbose > 1){
          fprintf(stderr, "discarding short or non[ip/arp] packet\n");
        }
        hr = 0;
        subre++;
      } else {
        sw = (0xff00 & (buffer[1] << 8)) | (buffer[0] & 0xff);
        if(hr >= (sw + 2)){
          if(verbose > 2){
            fprintf(stderr, "---> to tap: %d-(2+%d)\n", hr, sw);
            printbuffer(buffer + 2, sw, 70);
          }
          wr = write(tapfd, buffer + 2, sw);
          if(wr < sw){
            if(verbose){
              fprintf(stderr, "tap: failed write to tap: %s\n", strerror(errno));
            }
            run = 0;
            hr = 0;
          } else {
            subrc++;
            hr -= (sw + 2);
            if((hr > 0) && (rr == usbbuf)){
              memmove(buffer, buffer + sw + 2, hr );
            } else {
              hr = 0;
            }
          }
        }
      }
    } else {
      if(verbose){
        fprintf(stderr, "usb: read failed: %s (rr=%d) (packets send=%d, broken=%d)\n", (rr < 0) ? strerror(errno) : "eof", rr, subrc, subre);
      }
      if(rr < 0){
        switch(errno){
          case EAGAIN :
          case EINTR : 
          case ETIMEDOUT : 
            /* WARNING: there used to be code here to increase the timeout */
            hr = 0;
            break;
          default :
            break;
        }
      }
    }
  }

  if(tapfd >= 0){
    close(tapfd);
    tapfd = (-1);
  }

  return NULL;
}

#define CNX_FALSE        0x00
#define CNX_POLL         0x92
#define CNX_LOCATION     0x96
#define CNX_READSTOP     0x97
#define CNX_READRAM      0x98

#define CNE_NOTREADY     0x00
#define CNE_READY        0x01
#define CNE_ERRORDOWN    0x02
#define CNE_BLOCKUP      0x03
#define CNE_REBOOTDOWN   0x04
#define CNE_SHUTDOWN     0x05

#define STATE_ERROR        -1
#define STATE_END           0
#define STATE_IDLE          1
#define STATE_SAVE          2
#define STATE_REBOOT        3
#define STATE_READ          4
#define STATE_COMPLETE      5

void *subctl(void *data)
{
  unsigned char buffer[BUFFER];
  int rr, wr, state, reprieve, tries, timeout, ok;
  unsigned long base, need, have, got, pos, len;
  unsigned char *filename, *tmpname;
  FILE *fp;

  filename = (char *) data;
  if(filename){
    len = strlen(filename) + 16;
    tmpname = malloc(len);
    if(tmpname == NULL){
      if(verbose){
        fprintf(stderr, "ctl: unable to allocate buffer for %s\n", filename);
      }
      filename = NULL;
    } else {
      snprintf(tmpname, len - 1, "%s-%u", filename, getpid());
      tmpname[len - 1] = '\0';
    }
  }

  base = 0;
  need = 0;
  fp = NULL;

  for(state = STATE_ERROR; state != STATE_END; ){
    if(tapfd < 0){
      state = STATE_END;
    }

    for(reprieve = 0, timeout = TIMEOUT; timeout > 0;){
      rr = usb_bulk_read(usbah, usbctl | 0x80, buffer, usbbuf, timeout);
      if(verbose){
        fprintf(stderr, "ctl: read attempt: cmd=0x%02x, rr=%d, timeout=%d\n", buffer[0], rr, timeout);
      }
      if(rr > usbbuf){
        fprintf(stderr, "got too much %d\n", rr);
        exit(1);
      }
      if(rr < 0){
        fprintf(stderr, "ctl: read failed: %s\n", usb_strerror());
        timeout = 0;
        switch(errno){
          case EAGAIN :
          case EINTR : 
          case ETIMEDOUT : 
            break;
          default :
            state = STATE_ERROR;
            break;
        }
      } else if(rr == 0){
        timeout = (timeout > 20) ? 20 : reprieve;
        reprieve = 0;
      } else {
        if(timeout > 40){
          timeout = timeout / 4;
        } else {
          timeout--;
        }
        reprieve = timeout;
        if(verbose > 1){
          fprintf(stderr, "---> ctl read\n");
          printbuffer(buffer, rr, 70);
        }
        switch(buffer[0]){
          case CNX_POLL : 
            switch(buffer[4]){
              case CNE_NOTREADY     :
              case CNE_READY        :
                state = STATE_IDLE;
                sleep(1);
                break;
              case CNE_BLOCKUP      :
                state = STATE_SAVE;
                if(verbose){
                  fprintf(stderr, "ctl: *** block download ***\n");
                }
                break;
              case CNE_REBOOTDOWN   :
                state = STATE_REBOOT;
                if(verbose){
                  fprintf(stderr, "ctl: *** system reboot ***\n");
                }
                break;
              case CNE_ERRORDOWN    : /* ignored */
              case CNE_SHUTDOWN     :
                break;
              default :
                state = STATE_ERROR;
                break;
            }
            break;
          case CNX_LOCATION :
            if(state == STATE_READ){
              if(verbose){
                fprintf(stderr, "ctl: new size while downloading\n");
              }
            }
            if(fp){
              rewind(fp);
            } else {
              fp = fopen(tmpname, "w");
              if(fp == NULL){
                if(verbose){
                  fprintf(stderr, "ctl: unable to open %s: %s\n", tmpname, strerror(errno));
                }
              }
            }
            need = *((unsigned long *)(buffer + 4));
            base = *((unsigned long *)(buffer + 8));
            have = 0;
            if(verbose){
              fprintf(stderr, "ctl: need to get %lu@%lu\n", need, base);
            }
            if(need > 0){
              state = STATE_READ;
            } else {
              state = STATE_IDLE;
            }
            break;
          case CNX_READRAM :
            got = *((unsigned long *)(buffer + 4));
            pos = *((unsigned long *)(buffer + 8));
            ok = 1;

            if(got > (usbbuf - 12)){
              if(verbose){
                fprintf(stderr, "ctl: fragment too large %lu\n", got);
              }
              ok = 0;
            }

            if((pos > (base + have)) || ((pos < (base + have)) && ((pos + got) > (base + have)))){
              if(verbose){
                fprintf(stderr, "ctl: mismatched fragment %lu != %lu\n", pos, base + have);
              }
              ok = 0;
            } 

            if((base + have) != pos){
              if(verbose){
                fprintf(stderr, "ctl: stale fragment %lu@%lu\n", got, pos);
              }
              ok = 0;
            }

            if(ok){
              if(verbose){
                fprintf(stderr, "ctl: new fragment %lu@%lu [%lu/%lu]\n", got, pos, have, need);
              }
              if(fp){
                if(fwrite(buffer + 12, 1, got, fp) != got){
                  fprintf(stderr, "ctl: unable to save: %s\n", strerror(errno));
                }
              }
              have += got;
              if(have >= need){
                state = STATE_COMPLETE;
              } else {
                state = STATE_READ;
              }
            }
            break;
          case CNX_READSTOP :
            if(fp){
              if(rename(tmpname, filename)){
                if(verbose){
                  fprintf(stderr, "ctl: unable to replace %s: %s\n", filename, strerror(errno));
                }
              }
              fclose(fp);
              fp = NULL;
            }
            need = 0;
            state = STATE_IDLE;
            break;
          default :
            /* report error */
            if(verbose){
              fprintf(stderr, "ctl: *** unknown packet 0x%02x ***\n", buffer[0]);
            }
            state = STATE_ERROR;
            break;
        }
      }
    }

    if(verbose){
      fprintf(stderr, "ctl: finished reading, now writing (state=%d)\n", state);
    }

    memset(buffer, 0, usbbuf);
    switch(state){
      case STATE_ERROR :
        usb_clear_halt(usbah, usbctl);
        usb_clear_halt(usbah, 0x80 | usbctl);
        buffer[0] = CNX_FALSE;
        state = STATE_IDLE;
        break;
      case STATE_IDLE :
        buffer[0] = CNX_POLL;
        break;
      case STATE_SAVE : 
        buffer[0] = CNX_LOCATION;
        buffer[2] = 2;
        break;
      case STATE_REBOOT : 
        buffer[0] = CNX_LOCATION;
        buffer[2] = 0;
        break;
      case STATE_READ : 
        buffer[0] = CNX_READRAM;
        buffer[2] = 0;
        *((unsigned long *)(buffer + 4)) = ((need - have) > (usbbuf - 12)) ? (usbbuf - 12) : (need - have);
        *((unsigned long *)(buffer + 8)) = base + have;
        break;
      case STATE_COMPLETE :
        buffer[0] = CNX_READSTOP;
        break;
      default :
        buffer[0] = CNX_FALSE;
        break;
    }

    if(verbose){
      fprintf(stderr, "ctl: %s writing cmd=0x%02x\n", (buffer[0] == CNX_FALSE) ? "not" : "", buffer[0]);
    }

    if(buffer[0] != CNX_FALSE){
      for(tries = 0; tries < RESEND; tries++){
        wr = usb_bulk_write(usbah, usbctl, buffer, usbbuf, timeout);
        if(wr >= usbbuf){
          tries = RESEND;
        }
        if(verbose){
          fprintf(stderr, "ctl: write wr=%d, cmd=0x%02x %s\n", wr, buffer[0], (wr >= 0) ? "ok" : usb_strerror());
        }
      }
    }

  }

  if(tmpname){
    free(tmpname);
  }
  if(fp){
    fclose(fp);
  }

  close(tapfd);
  tapfd = (-1);

  return NULL;
}

void usage()
{
  printf("usage: tap2usb [-x] [-v] [-i] [-d] [-t tap name] [-m manufacturer] [-p product] [-n interface] [-f configuration] [-b usb buffer] [-e ethernet endpoint] [-a control endpoint] [-y filesystem]\n");
}

int main(int argc, char **argv)
{
  unsigned char name[NAME];
  char *tmp;
  int i, j;
  int value, flag;
  char *tap;
  int interactive, monitored;
  int vendor, product, inter, config;
  pthread_t tousb, totap, onctl;
  char *filesystem;

#if 0
  int rxf, txf, rxb, txb;
  int ert, ersw, ersr;
#endif

  i = 1;
  j = 1;
  tap = NULL;
  filesystem = "filesystem";

  usbctl  = TAPUSB_CTLEP; /* in = out | 0x80 */
  usbeth  = TAPUSB_ETHEP; /* in = out | 0x80 */
  vendor  = TAPUSB_VENDOR;
  product = TAPUSB_PRODUCT;
  config  = TAPUSB_CONFIG;
  inter   = TAPUSB_INTERFACE;
  usbbuf  = TAPUSB_BUFFER;

  interactive = isatty(STDIN_FILENO);
  monitored = 1;
  
  while(i < argc){
    if(argv[i][0] == '-'){
      flag = argv[i][j];
      switch(flag){
        case 'y' :
          j++;
          if(argv[i][j] == '\0'){
            i++;
            j = 0;
          }
          filesystem = argv[i] + j;
          if(filesystem == NULL){
            fprintf(stderr, "option -y needs a filename\n");
            return EX_USAGE;
          }
          i++;
          j = 1;
          break;
        case 'm' :
        case 'p' :
        case 'n' :
        case 'f' :
        case 't' :
        case 'e' :
        case 'a' :
        case 'b' :
          j++;
          if(argv[i][j] == '\0'){
            i++;
            j = 0;
          }
          tmp = argv[i] + j;
          if(tmp == NULL){
            fprintf(stderr, "option -%c needs an argument\n", flag);
            return EX_USAGE;
          }
          value = atoi(tmp);
          switch(flag){
            case 'p' :
              product = value;
              break;
            case 'm' :
              vendor = value;
              break;
            case 'n' :
              inter = value;
              break;
            case 'f' :
              config = value;
              break;
            case 'e' :
              usbeth = value;
              break;
            case 'a' :
              usbctl = value;
              break;
            case 'b' :
              if((value <= 0) || (value > BUFFER)){
                fprintf(stderr, "invalid buffer size %d\n", value);
                return EX_USAGE;
              }
              usbbuf = value;
              break;
            case 't' :
              tap = tmp;
              break;
          }
          i++;
          j = 1;
          break;
        case '\0' :
          i++;
          j = 1;
          break;
        case 'h' :
          usage();
          return EX_OK;
          break;
        case 'v' :
          j++;
          verbose++;
          break;
        case 'x' :
          j++;
          monitored = 0;
          break;
        case 'd' :
          j++;
          interactive = 0;
          break;
        case 'i' :
          j++;
          interactive = 1;
          break;
        default :
          fprintf(stderr, "unknown option %c\n", argv[i][1]);
          return EX_USAGE;
          break;
      }
    } else {
      fprintf(stderr, "argument syntax error\n");
      return EX_USAGE;
    }
  }

  signal(SIGPIPE, SIG_IGN);

  if(tap == NULL){
    snprintf(name, NAME - 1, "usb%d", vendor & 0xf);
    name[NAME - 1] = '\0';
    tap = name;
  }

  tapfd = tap_open(tap);
  if(tapfd < 0){
    if(verbose){
      fprintf(stderr, "unable to open tap device %s\n", tap);
    }
    return EX_OSERR;
  }

  usbah = sub_open(vendor, product, config, inter);
  if(usbah == NULL){
    if(verbose){
      fprintf(stderr, "unable to locate or open usb device 0x%x:0x%x\n", vendor, product);
    }
    return EX_OSERR;
  }
  printf("point 1\n");
  if(verbose){
    fprintf(stderr, "tap=%s, usb=0x%x:0x%x\n", tap, vendor, product);
  }
  
  signal(SIGINT, handle);
  printf("point 2 \n");
  signal(SIGTERM, handle);

  /* Confession: I dislike threads, but it seemed the least evil option 
   * given that libusb blocks in read/writes. It seems we can get away 
   * without locking though that should be fixed on principle, as should the 
   * general uglyness
   */

  if(interactive == 0){
    if(daemon(0, 0)){
      fprintf(stderr, "unable to enter background mode: %s\n", strerror(errno));
    }
    printf("point 3\n");
    syslog(LOG_NOTICE | LOG_DAEMON, "tap2usb: starting tap to usb bridge");
    verbose = 0;
  }

  if(monitored){
    if(pthread_create(&onctl, NULL, subctl, filesystem)){
      fprintf(stderr, "unable to start control thread\n");
      return EX_OSERR;
    }
  }

  if(pthread_create(&totap, NULL, sub2tap, NULL)){
    fprintf(stderr, "unable to start receive thread\n");
    return EX_OSERR;
  }
  if(pthread_create(&tousb, NULL, tap2sub, NULL)){
    fprintf(stderr, "unable to start send thread\n");
    return EX_OSERR;
  }

  if(interactive){
    console();
  }

  pthread_join(tousb, NULL);
  pthread_cancel(totap);
  pthread_join(totap, NULL);

  if(monitored){
    pthread_cancel(onctl);
    pthread_join(onctl, NULL);
  }

  sub_close(usbah, inter);
  tap_close(tapfd);

  if(interactive == 0){
    syslog(LOG_NOTICE | LOG_DAEMON, "tap2usb: error rate: read=%d/%d write=%d/%d", subre, subrc, subwe, subwc);
    syslog(LOG_NOTICE | LOG_DAEMON, "tap2usb: stopped tap to usb bridge");
  }

  return 0;
}
